-- Machine Level Access
include machine.e
include _common_.e
include _types_.e

-- char *
--   addr = allocate_string(s)
--   s = peek_string(addr)

-- wchar *
--   addr = allocate_wstring(s)
--   s = peek_wstring(addr)

-- char **
--   addr = allocate_string_list(s)
--   s = peek_string_list( addr, max )

--------------------------------------------------------------------------------

global function peeks(object address)
--<function>
--<name>peeks</name>
--<digest>read one or more signed bytes from an address in memory</digest>
--<desc>
--</desc>
--<param>
--<type>object</type>
--<name>address</name>
--<desc>
-- either of
-- * an atom addr : to fetch one byte at addr, or
-- * a pair {addr,len} : to fetch len bytes at addr
--</desc>
--</param>
--<return>
--</return>
--<example>
-- poke(addr, {-1,1,2,3})
-- ? peeks(addr) --> -1
-- ? peeks({addr, 4}) --> {-1, 1, 2, 3}
--</example>
--<see_also>poke(), peek()</see_also>
--</function>
  integer n
  sequence s

  if atom(address) then
    n = peek(address)
    if n >= #80 then
      return n - #100
    else
      return n
    end if
  else
    s = {}
    for i = 1 to address[2] do
      s = append(s, peeks(address[1]+i-1))
    end for
    return s
  end if
end function

--------------------------------------------------------------------------------

global procedure poke2(atom address, object i)
--<procedure>
--<name>poke2</name>
--<digest>put one or more 2-byte signed/unsigned words at a memory location</digest>
--<desc></desc>
--<param>
--<type>atom</type>
--<name>address</name>
--<desc>either a word or a non empty sequence of words</desc>
--</param>
--<param>
--<type>integer</type>
--<name>i</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- poke2(addr, {-1, 4096, 400})
--</example>
--<see_also>peek2u(), peek2s()</see_also>
--</procedure>
  integer l, h

  if atom(i) then
    if (i > #10000) then return end if
    l = remainder(i, #100)
    h = floor(i / #100)
    poke(address, {l, h})
  else
    for n = 1 to length(i) do
      poke2(address+2*(n-1), i[n])
    end for
  end if
end procedure

--------------------------------------------------------------------------------

global function peek2u(object address)
--<function>
--<name>peek2u</name>
--<digest>read one or more 2-byte unsigned words from an address in memory</digest>
--<desc></desc>
--<param>
--<type>object</type>
--<name>address</name>
--<desc>
-- either of
-- * an atom addr : to fetch one word at addr, or
-- * a pair {addr,len} : to fetch len words at addr
--</desc>
--</param>
--<return>
--</return>
--<example>
-- poke2(addr, {-1, 4096, 400})
-- ? peek2u(addr) --> 65535
-- ? peek2u({addr, 3}) --> {65535,4096,400}
--</example>
--<see_also>poke2(), peek2s()</see_also>
--</function>
  sequence s

  if atom(address) then
    s = peek({address, 2})
    return  s[1] + s[2]*#100
  else
    s = {}
    for i = 1 to address[2] do
      s = append(s, peek2u(address[1]+2*(i-1)))
    end for
    return s
  end if
end function

--------------------------------------------------------------------------------

global function peek2s(object address)
--<function>
--<name>peek2s</name>
--<digest>read one or more 2-byte signed words from an address in memory</digest>
--<desc></desc>
--<param>
--<type>object</type>
--<name>address</name>
--<desc>
-- either of
-- * an atom addr : to fetch one word at addr, or
-- * a pair {addr,len} : to fetch len words at addr
--</desc>
--</param>
--<return>
--</return>
--<example>
-- poke2(addr, {-1, 4096, 400})
-- ? peek2s(addr) --> -1
-- ? peek2s({addr, 3}) --> {-1,4096,400}
--</example>
--<see_also>poke2(), peek2u()</see_also>
--</function>
  integer n
  sequence s

  if atom(address) then
    s = peek({address, 2})
    n = s[1] + s[2]*#100
    if n >= #8000 then
      return n - #10000
    else
      return n
    end if
  else
    s = {}
    for i = 1 to address[2] do
      s = append(s, peek2s(address[1]+2*(i-1)))
    end for
    return s
  end if
end function

--------------------------------------------------------------------------------

global procedure poke8(atom address, object x)
--<procedure>
--<name>poke8</name>
--<digest>put one or more 8-byte signed/unsigned long integers at a memory location</digest>
--<desc></desc>
--<param>
--<type>atom</type>
--<name>address</name>
--<desc>either a long integer or a non empty sequence of long integers</desc>
--</param>
--<param>
--<type>object</type>
--<name>x</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- poke8(addr, -1)
--</example>
--<see_also>peek8u(), peek8s()</see_also>
--</procedure>
  if atom(x) then x = {x} end if
  address -= 8
  for i = 1 to length(x) do
    -- put less significant bits.
    poke4(address+i*8,and_bits(#FFFFFFFF,x[i]))
    if x[i] > 0 then
      poke4(address+i*8+4,floor(x[i]/#100000000))
    else
      poke4(address+i*8+4,floor(x[i]/#100000000))
    end if
  end for
end procedure

--------------------------------------------------------------------------------

global function peek8u(object address)
--<function>
--<name>peek8u</name>
--<digest>read one or more 8-byte unsigned long integers from an address in memory</digest>
--<desc></desc>
--<param>
--<type>object</type>
--<name>address</name>
--<desc>
-- either of
-- * an atom addr : to fetch one long integer at addr, or
-- * a pair {addr,len} : to fetch len long integers at addr
--</desc>
--</param>
--<return>
--</return>
--<example>
-- poke8(addr, -1)
-- ? peek8u(addr) --> 1.844674407e+019
--</example>
--<see_also>poke8(), peek8s()</see_also>
--</function>
  integer count, atom_flag
  sequence out

  count = 1
  atom_flag = atom(address)
  if sequence(address) then
    count = address[2]
    address = address[1]
  end if
  out = peek4u({address,count*2})
  for i = 1 to length(out) by 2 do
    out[floor((i+1)/2)] = out[i+1] * #100000000 + out[i]
  end for
  out = out[1..count]
  if atom_flag then return out[1] end if
  return out
end function

--------------------------------------------------------------------------------

global function peek8s(object address)
--<function>
--<name>peek8s</name>
--<digest>read one or more 8-byte signed long integers from an address in memory</digest>
--<desc>
--</desc>
--<param>
--<type>object</type>
--<name>address</name>
--<desc>
-- either of
-- * an atom addr : to fetch one long integer at addr, or
-- * a pair {addr,len} : to fetch len long integers at addr
--</desc>
--</param>
--<return>
--</return>
--<example>
-- poke8(addr, -1)
-- ? peek8s(addr) --> -1
--</example>
--<see_also>poke8(), peek8u()</see_also>
--</function>
  integer count, atom_flag
  sequence sout, uout, out

  count = 1
  atom_flag = atom(address)
  if sequence(address) then
    count = address[2]
    address = address[1]
  end if
  sout = peek4s({address,2*count})
  uout = peek4u({address,2*count})
  out = repeat(0,count)
  for i = 1 to count do
    out[i] = sout[2*i] * #100000000 + uout[2*i-1]
  end for
  if atom_flag then return out[1] end if
  return out
end function

--------------------------------------------------------------------------------

global procedure poke_float(atom p, atom a)
--<procedure>
--<name>poke_float</name>
--<digest>put a 4-byte float at a memory location</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>p</name>
--<desc></desc>
--</param>
--<param>
--<type>atom</type>
--<name>a</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- poke_float(addr, 3.1416)
--</example>
--<see_also>peek_float()</see_also>
--</procedure>
  poke(p, atom_to_float32(a))
end procedure

--------------------------------------------------------------------------------

global function peek_float(atom p)
--<function>
--<name>peek_float</name>
--<digest>read a 4-byte float from an address in memory</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>p</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- poke_float(addr, 3.1416)
-- ? peek_float(addr) --> 3.141599894 (caused by float storage format)
--</example>
--<see_also>poke_float()</see_also>
--</function>
  return float32_to_atom(peek({p, 4}))
end function

--------------------------------------------------------------------------------

global procedure poke_double(atom p, atom a)
--<procedure>
--<name>poke_double</name>
--<digest>put a 8-byte float at a memory location</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>p</name>
--<desc></desc>
--</param>
--<param>
--<type>atom</type>
--<name>a</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- poke_double(addr, 216656332.2)
--</example>
--<see_also>peek_double()</see_also>
--</procedure>
  poke(p, atom_to_float64(a))
end procedure

--------------------------------------------------------------------------------

global function peek_double(atom p)
--<function>
--<name>peek_double</name>
--<digest>read a 8-byte float from an address in memory</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>p</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- poke_double(addr, 216656332.2)
-- ? peek_double(addr) --> 216656332.2
--</example>
--<see_also>poke_double()</see_also>
--</function>
  return float64_to_atom(peek({p, 8}))
end function

--------------------------------------------------------------------------------

global procedure poke_pointer(atom addr, object x)
--<procedure>
--<name>poke_pointer</name>
--<digest>put a 4 byte pointer in memory</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>addr</name>
--<desc></desc>
--</param>
--<param>
--<type>object</type>
--<name>x</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- poke_pointer(addr, allocate_string("where?"))
--</example>
--<see_also>peek_pointer()</see_also>
--</procedure>
  if addr <= 0 then
    error_message(sprintf("poke_pointer: bad pointer %d\n", addr), 1)
  end if
  poke4(addr, x)
end procedure

--------------------------------------------------------------------------------

global function peek_pointer(object addr)
--<function>
--<name>peek_pointer</name>
--<digest>read a 4 byte pointer in memory</digest>
--<desc>
--</desc>
--<param>
--<type>object</type>
--<name>addr</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- addr2 = allocate_string("where?")
-- ? addr2
-- poke_pointer(addr, addr2)
-- ? peek_pointer(addr) --> same as addr2
--</example>
--<see_also>poke_pointer()</see_also>
--</function>
  if atom(addr) and (addr <= 0) then
    error_message(sprintf("peek_pointer: bad pointer %d\n", addr), 1)
  end if
  if sequence(addr) and (addr[1] <= 0) then
    error_message(sprintf("peek_pointer: bad pointer %d\n", addr[1]), 1)
  end if
  return peek4u(addr)
end function

--------------------------------------------------------------------------------

global procedure poke_string(atom addr, integer size, sequence s)
--<procedure>
--<name>poke_string</name>
--<digest>put a string in memory</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>addr</name>
--<desc></desc>
--</param>
--<param>
--<type>integer</type>
--<name>size</name>
--<desc></desc>
--</param>
--<param>
--<type>sequence</type>
--<name>s</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- poke_string(addr2, 4, "here")
--  --> poke_string fails because 4 is not enough to include terminal 0
-- poke_string(addr2, 5, "here")
--  --> poke_string is successful
--</example>
--<see_also>peek_string()</see_also>
--</procedure>
  integer len

  if addr <= 0 then return end if
  len = length(s)
  if size <= len then return end if
  poke(addr, s)
  poke(addr + len, 0)
end procedure

--------------------------------------------------------------------------------

global function peek_string(atom addr)
--<function>
--<name>peek_string</name>
--<digest>read a string in memory</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>addr</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- poke_string(addr2, 5, "here")
-- puts(1, peek_string(addr2) & "\n") --> here
--</example>
--<see_also>poke_string()</see_also>
--</function>
  atom ptr

  ptr = addr
  while peek(ptr) do ptr += 1 end while
  return peek({addr, ptr-addr})
end function

--------------------------------------------------------------------------------

global function allocate_wstring(sequence s)
--<function>
--<name>allocate_wstring</name>
--<digest>allocate a wide string in memory</digest>
--<desc>
--</desc>
--<param>
--<type>sequence</type>
--<name>s</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- addr2 = allocate_wstring("lgant")
--</example>
--<see_also>poke_wstring(), peek_wstring()</see_also>
--</function>
  atom addr
  integer len

  len = length(s) * 2
  addr = allocate(len + 2)
  if addr then
    poke2(addr, s)
    poke2(addr + len, 0)
  end if
  return addr
end function

--------------------------------------------------------------------------------

global procedure poke_wstring(atom addr, integer size, sequence s)
--<procedure>
--<name>poke_wstring</name>
--<digest>puts a wide string in memory</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>addr</name>
--<desc></desc>
--</param>
--<param>
--<type>integer</type>
--<name>size</name>
--<desc></desc>
--</param>
--<param>
--<type>sequence</type>
--<name>s</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- poke_wstring(addr2, 10, "ruin")
--  --> poke_wstring fails because 10 is not enough to include terminal 0
-- poke_wstring(addr2, 12, "ruin")
--  --> poke_wstring is successful
--</example>
--<see_also>allocate_wstring(), peek_wstring()</see_also>
--</procedure>
  integer len

  if addr <= 0 then return end if
  len = length(s) * 2
  if size <= len then return end if
  poke2(addr, s)
  poke2(addr + len, 0)
end procedure

--------------------------------------------------------------------------------

global function peek_wstring(atom addr)
--<function>
--<name>peek_wstring</name>
--<digest>reads a wide string in memory</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>addr</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- poke_wstring(addr2, 12, "ruin")
-- puts(1, peek_wstring(addr2) & "\n")  --> ruin
--</example>
--<see_also>allocate_wstring(), poke_wstring()</see_also>
--</function>
  atom ptr

  ptr = addr
  while peek2u(ptr) do ptr += 2 end while
  return peek2u({addr, (ptr - addr) / 2})
end function

--------------------------------------------------------------------------------

global function allocate_plist(sequence pointers)
--<function>
--<name>allocate_plist</name>
--<digest>allocates a list of pointers in memory</digest>
--<desc>
--</desc>
--<param>
--<type>sequence</type>
--<name>pointers</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- addr2 = allocate(4)
-- addr3 = allocate(4)
-- addr = allocate_plist({addr2, addr3})
--</example>
--<see_also>poke_plist(), peek_plist(), free_plist()</see_also>
--</function>
  atom addr
  integer len

  len = length(pointers) * address_length
  addr = allocate(len + address_length)
  if addr then
    poke_pointer(addr, pointers)
    poke_pointer(addr + len, NULL)
  end if
  return addr
end function

--------------------------------------------------------------------------------

global procedure poke_plist(atom addr, sequence pointers)
--<procedure>
--<name>poke_plist</name>
--<digest>puts a list of pointers in memory</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>addr</name>
--<desc></desc>
--</param>
--<param>
--<type>sequence</type>
--<name>pointers</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- addr2 = allocate(4)
-- addr3 = allocate(4)
-- addr = allocate_plist({NULL, NULL})
-- poke_plist(addr, {addr3, addr2})
--</example>
--<see_also>allocate_plist(), peek_plist(), free_plist()</see_also>
--</procedure>
  integer len

  if addr <= 0 then return end if
  len = length(pointers) * address_length
  poke_pointer(addr, pointers)
  poke_pointer(addr + len, NULL)
end procedure

--------------------------------------------------------------------------------

global function peek_plist(atom addr)
--<function>
--<name>peek_plist</name>
--<digest>reads a list of pointers in memory</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>addr</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- addr2 = allocate(4)
-- ? addr2 --> 4758384
-- addr3 = allocate(4)
-- ? addr3 --> 4832984
-- addr = allocate_plist({addr2, addr3})
-- ? peek_plist(addr) --> {4758384,4832984}
--</example>
--<see_also>allocate_plist(), poke_plist(), free_plist()</see_also>
--</function>
  atom ptr

  ptr = addr
  while peek_pointer(ptr) do ptr += address_length end while
  return peek_pointer({addr, (ptr - addr) / address_length})
end function

--------------------------------------------------------------------------------

global procedure free_plist(atom addr)
--<procedure>
--<name>free_plist</name>
--<digest>free a list of pointers in memory</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>addr</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- free_plist(addr)
-- ? peek_plist(addr) --> {}
--</example>
--<see_also>allocate_plist(), poke_plist(), peek_plist()</see_also>
--</procedure>
  atom ptr

  ptr = peek_pointer(addr)
  while ptr do
    free(ptr)
    poke4(addr, 0)
    addr += address_length
    ptr = peek_pointer(addr)
  end while
end procedure

--------------------------------------------------------------------------------

global function allocate_string_list(sequence string_list)
--<function>
--<name>allocate_string_list</name>
--<digest>allocate a list of strings in memory</digest>
--<desc>
-- no need of poke_string_list as allocate_string_list does the job
--</desc>
--<param>
--<type>object</type>
--<name>string_list</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- addr = allocate_string_list({"my","taylor","is","rich"})
--</example>
--<see_also>peek_string_list(), free_string_list()</see_also>
--</function>
  for i = 1 to length(string_list) do
    string_list[i] = allocate_string(string_list[i])
  end for
  return allocate_plist(string_list)
end function

--------------------------------------------------------------------------------

global function peek_string_list( atom addr, sequence optional)
--<function>
--<name>peek_string_list</name>
--<digest>reads a list of strings in memory</digest>
--<desc>
-- opposite of allocate_string_list()
--</desc>
--<param>
--<type>atom</type>
--<name>addr</name>
--<desc></desc>
--</param>
--<param>
--<type>sequence</type>
--<name>optional</name>
--<desc>
-- optional "max" fixes maximum number of strings to read
--</desc>
--</param>
--<return>
--</return>
--<example>
-- addr = allocate_string_list({"my","taylor","is","rich"})
-- s = peek_string_list(addr, {})
-- s is {"my","taylor","is","rich"}
-- s = peek_string_list(addr, {{"max", 2}})
-- s is {"my","taylor"}
--</example>
--<see_also>allocate_string_list(), free_string_list()</see_also>
--</function>
  sequence strings, str
  atom ptr
  integer max

  strings = {}
  max = get_option("max", optional, -1)
  ptr = peek_pointer(addr)
  while ptr do
    str = peek_string( ptr )
    strings = append( strings, str )
    if (max != -1) and (length(strings) = max) then
      exit
    end if
    addr += address_length
    ptr = peek_pointer(addr)
  end while
  return strings
end function

--------------------------------------------------------------------------------

global procedure free_string_list(atom addr)
--<procedure>
--<name>free_string_list</name>
--<digest>frees a list of strings in memory</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>addr</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- free_string_list(addr)
-- s = peek_string_list(addr, {})
-- s is {}
--</example>
--<see_also>allocate_string_list(), peek_string_list()</see_also>
--</procedure>
  free_plist(addr)
end procedure

--------------------------------------------------------------------------------

global function allocate_structure(atom p, sequence def)
--<function>
--<name>allocate_structure</name>
--<digest></digest>
--<desc>
-- allocates memory according to a structure definition
-- automatically sizes structure to EU version and OS Architecture
--</desc>
--<param>
--<type>atom</type>
--<name>p</name>
--<desc></desc>
--</param>
--<param>
--<type>sequence</type>
--<name>def</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- atom p = NULL
-- p = allocateStructure(p, {pointer, int, int} )
--</example>
--<see_also>
--</see_also>
--</function>
  integer lg
  atom ctype

--  log_printf("allocate_structure(#%x, %s)\n", {p, hex_sequence(def)})
  lg = 0
  for i = 1 to length(def) do
    ctype = convert_c_type(def[i])
    lg += size_of(ctype)
--    log_printf("allocate_structure: lg = %d\n", {lg})
  end for
  if lg then p = allocate(lg) end if
--    log_printf("allocate_structure: p = #%x, lg = %d\n", {p, lg})
  return p
end function

-----------------------------------------------------------------------------

global procedure write_structure(atom p, sequence def, sequence values)
--<procedure>
--<name>write_structure</name>
--<digest></digest>
--<desc>
-- writes data in memory according to a structure definition
-- automatically sizes structure to EU version and OS Architecture
--</desc>
--<param>
--<type>atom</type>
--<name>p</name>
--<desc></desc>
--</param>
--<param>
--<type>sequence</type>
--<name>def</name>
--<desc></desc>
--</param>
--<param>
--<type>sequence</type>
--<name>values</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- writeStructure(p, {pointer, int, int}, {hwnd, 12, 25} )
--</example>
--<see_also>
--</see_also>
--</procedure>
  atom ctype --, ptr

--  ptr = p
--  log_printf("write_structure(#%x, %s, %s)\n", {p, hex_sequence(def), sprint(values)})
  for i = 1 to length(def) do
    ctype = convert_c_type(def[i])
    if (ctype = int8) or (ctype = uint8) then  -- 8-bit signed, unsigned
      poke(p, values[i])
    elsif (ctype = int16) or (ctype = uint16) then  -- 16-bit signed, unsigned
      poke2(p, values[i])
    elsif (ctype = int32) or (ctype = uint32) then  -- 32-bit signed, unsigned
      poke4(p, values[i])
    elsif (ctype = int64) or (ctype = uint64) then  -- 64-bit signed, unsigned
      poke8(p, values[i])
    elsif ctype = pointer then       -- pointer
      if address_length=4 then     -- 32-bit
        poke4(p, values[i])
      elsif address_length=8 then  -- 64-bit
        poke8(p, values[i])
      end if
    elsif ctype = long_pointer then       -- 64-bit integer
      poke8(p, values[i])
    elsif ctype = float then       -- 32-bit float
      poke_float(p, values[i])
    elsif ctype = double then       -- 64-bit float
      poke_double(p, values[i])
    end if
    p += size_of(ctype)
  end for
--  log_printf("write_structure: p = #%x, code = %s\n", {ptr, hex_sequence(peek({ptr, 12}))})
end procedure

-----------------------------------------------------------------------------

global function read_structure(atom p, sequence def)
--<function>
--<name>read_structure</name>
--<digest></digest>
--<desc>
-- reads data in memory according to a structure definition
-- automatically sizes structure to EU version and OS Architecture
--</desc>
--<param>
--<type>atom</type>
--<name>p</name>
--<desc></desc>
--</param>
--<param>
--<type>sequence</type>
--<name>def</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- sequence s = readStructure(p, {pointer, int, int} )
--</example>
--<see_also>
--</see_also>
--</function>
  atom ctype
  sequence result

--  log_printf("read_structure(#%x, %s)\n", {p, hex_sequence(def)})
--  log_printf("read_structure: p = #%x, code = %s\n", {p, hex_sequence(peek({p, 12}))})
  result = {}
  for i = 1 to length(def) do
    ctype = convert_c_type(def[i])
    if ctype = int8 then       -- 8-bit signed
      result = append( result, peeks(p) )
    elsif ctype = int16 then       -- 16-bit signed
      result = append( result, peek2s(p) )
    elsif ctype = int32 then       -- 32-bit signed
      result = append( result, peek4s(p) )
    elsif ctype = int64 then       -- 64-bit signed
      result = append( result, peek8s(p) )
    elsif ctype = uint8 then       -- 8-bit unsigned
      result = append( result, peek(p) )
    elsif ctype = uint16 then       -- 16-bit unsigned
      result = append( result, peek2u(p) )
    elsif ctype = uint32 then       -- 32-bit unsigned
      result = append( result, peek4u(p) )
    elsif ctype = uint64 then       -- 64-bit unsigned
      result = append( result, peek8u(p) )
    elsif ctype = pointer then       -- pointer
      if address_length=4 then     -- 32-bit
        result = append( result, peek4s(p) )
      elsif address_length=8 then  -- 64-bit
        result = append( result, peek8s(p) )
      end if
    elsif ctype = long_pointer then       -- 64-bit integer
      result = append( result, peek8u(p) )
    elsif ctype = float then       -- 32-bit float
      result = append( result, peek_float(p) )
    elsif ctype = double then       -- 64-bit float
      result = append( result, peek_double(p) )
    end if
    p += size_of(ctype)
  end for
  return result
end function

-----------------------------------------------------------------------------

global procedure free_structure(atom p, sequence def, sequence values)
--<procedure>
--<name>free_structure</name>
--<digest></digest>
--<desc>
-- frees memory according to a structure definition
-- automatically frees pointers included in the structure
-- according to EU version and OS Architecture
--</desc>
--<param>
--<type>atom</type>
--<name>p</name>
--<desc></desc>
--</param>
--<param>
--<type>sequence</type>
--<name>def</name>
--<desc></desc>
--</param>
--<param>
--<type>sequence</type>
--<name>values</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- freeStructure(p, {pointer, int, int}, {hwnd, 12, 25} )
--</example>
--<see_also>
--</see_also>
--</procedure>
--  log_printf("free_structure(#%x, %s)\n", {p, hex_sequence(def)})
  for i = 1 to length(def) do
    if find(def[i], {pointer, long_pointer}) then
      free(values[i])
    end if
  end for
  free(p)
end procedure

--------------------------------------------------------------------------------

global function cdecl_callback(integer rtn)
--<function>
--<name>cdecl_callback</name>
--<digest>return a CDECL call_back handle for a routine</digest>
--<desc>
--</desc>
--<param>
--<type>integer</type>
--<name>rtn</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
--</example>
--<see_also>
--</see_also>
--</function>
if platform() = WIN32 then
  return call_back({ '+', rtn })
elsif platform() = LINUX then
  return call_back(rtn)
end if
end function

-------------------------------------------------------------------------------

global function byte_to_word(atom l, atom h)
--<function>
--<name>byte_to_word</name>
--<digest>merges 2 bytes in a word</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>l</name>
--<desc></desc>
--</param>
--<param>
--<type>atom</type>
--<name>h</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- ? byte_to_word(#AA, #80)  --> 32938 (#80AA)
--</example>
--<see_also>high_byte(), low_byte()</see_also>
--</function>
  return h*#100+l
end function

-------------------------------------------------------------------------------

global function high_byte(atom x)
--<function>
--<name>high_byte</name>
--<digest>extracts higher byte of a word</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>x</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- ? high_byte(#80AA)  --> 128 (#80)
--</example>
--<see_also>byte_to_word(), low_byte()</see_also>
--</function>
  return floor(x / #100)
end function

-------------------------------------------------------------------------------

global function low_byte(atom x)
--<function>
--<name>low_byte</name>
--<digest>extracts lower byte of a word</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>x</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- ? low_byte(#80AA)  --> 170 (#AA)
--</example>
--<see_also>byte_to_word(), high_byte()</see_also>
--</function>
  return remainder(x, #100)
end function

-------------------------------------------------------------------------------

global function word_to_dword(atom l, atom h)
--<function>
--<name>word_to_dword</name>
--<digest>merges 2 words in a double word</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>l</name>
--<desc></desc>
--</param>
--<param>
--<type>atom</type>
--<name>h</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- ? word_to_dword(#21FA, #8123)  --> 2166563322 (#812321FA)
--</example>
--<see_also>high_word(), low_word()</see_also>
--</function>
  return h*#10000+l
end function

--------------------------------------------------------------------------------

global function high_word(atom pData)
--<function>
--<name>high_word</name>
--<digest>extracts higher word of a double word</digest>
--<desc>
-- from win32lib tk_maths.e
--</desc>
--<param>
--<type>atom</type>
--<name>pData</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- ? high_word(#812321FA)  --> 33059 (#8123)
--</example>
--<see_also>word_to_dword(), low_word()</see_also>
--</function>
  return and_bits(and_bits(pData, #FFFF0000) / #10000, #FFFF)
end function

--------------------------------------------------------------------------------

global function low_word(atom pData)
--<function>
--<name>low_word</name>
--<digest>extracts lower word of a double word</digest>
--<desc>
-- from win32lib tk_maths.e
--</desc>
--<param>
--<type>atom</type>
--<name>pData</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- ? low_word(#812321FA)  --> 8698 (#21FA)
--</example>
--<see_also>word_to_dword(), high_word()</see_also>
--</function>
  return and_bits(pData, #FFFF)
end function

-------------------------------------------------------------------------------

global function higher_bits(integer x, integer divide)
--<function>
--<name>higher_bits</name>
--<digest>returns the highest bits of an integer</digest>
--<desc>
--</desc>
--<param>
--<type>integer</type>
--<name>x</name>
--<desc></desc>
--</param>
--<param>
--<type>integer</type>
--<name>divide</name>
--<desc>divide is the is the value that fixes the limit between higher bits and lower bits</desc>
--</param>
--<return>
--</return>
--<example>
--  higher_bits(x, #10) gets the higher quartet (4 bits) of byte x (8 bits)
--  higher_bits(x, #100) gets the higher bytes (8 bits) of word x (16 bits)
--  higher_bits(x, #10000) gets the higher word (16 bits) of dword x (32 bits)
--</example>
--<see_also></see_also>
--</function>
  return floor(x / divide)
end function

-------------------------------------------------------------------------------

global function lower_bits(integer x, integer divide)
--<function>
--<name>lower_bits</name>
--<digest>returns the lowest bits of an integer</digest>
--<desc>
--</desc>
--<param>
--<type>integer</type>
--<name>x</name>
--<desc></desc>
--</param>
--<param>
--<type>integer</type>
--<name>divide</name>
--<desc>divide is the is the value that fixes the limit between higher bits and lower bits</desc>
--</param>
--<return>
--</return>
--<example>
--  lower_bits(x, #10) gets the lower quartet (4 bits) of byte x (8 bits)
--  lower_bits(x, #100) gets the lower bytes (8 bits) of word x (16 bits)
--  lower_bits(x, #10000) gets the lower word (16 bits) of dword x (32 bits)
--</example>
--<see_also></see_also>
--</function>
  return remainder(x, divide)
end function

-------------------------------------------------------------------------------

global function join_bits(integer highest, integer lowest, integer divide)
--<function>
--<name>join_bits</name>
--<digest>combine higher and lower item to build one bigger item</digest>
--<desc>
--</desc>
--<param>
--<type>integer</type>
--<name>highest</name>
--<desc></desc>
--</param>
--<param>
--<type>integer</type>
--<name>lowest</name>
--<desc></desc>
--</param>
--<param>
--<type>integer</type>
--<name>divide</name>
--<desc>divide is the resulting value when highest is 1 and lowest is 0</desc>
--</param>
--<return>
--</return>
--<example>
--  join_bits(h, l, #10) joins to quartets (4 bits) to build a byte (8 bits)
--  join_bits(h, l, #100) joins to bytes (8 bits) to build a word (16 bits)
--  join_bits(h, l, #10000) joins to words (16 bits) to build a dword (32 bits)
--</example>
--<see_also></see_also>
--</function>
  return (highest*divide) + lowest
end function

-------------------------------------------------------------------------------

global function or_all(object a)
--<function>
--<name>or_all</name>
--<digest>Or's together all atoms in the argument</digest>
--<desc>
--</desc>
--<param>
--<type>object</type>
--<name>a</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- ? or_all({1,2,4,8})  --> 15
--</example>
--<see_also>
--</see_also>
--</function>
  atom b

  if atom(a) then return a end if
  b = 0
  for i = 1 to length(a) do
    if atom(a[i]) then
      b = or_bits(b, a[i])
    else
      b = or_bits(b, or_all(a[i]))
    end if
  end for
  return b
end function

-------------------------------------------------------------------------------

global function shift_bits(object source_number, integer shift_distance)
--<function>
--<name>shift_bits</name>
--<digest>Moves the bits in the input value by the specified distance.</digest>
--<desc>
--</desc>
--<param>
--<type>object</type>
--<name>source_number</name>
--<desc></desc>
--</param>
--<param>
--<type>integer</type>
--<name>shift_distance</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- printf(1, "%08x\n", shift_bits(#ABCD,  4))  --> 00000ABC
-- printf(1, "%08x\n", shift_bits(#ABCD, -4))  --> 000ABCD0
--</example>
--<see_also>rotate_bits()</see_also>
--</function>
  integer lSigned

  if sequence(source_number) then
    for i = 1 to length(source_number) do
      source_number[i] = shift_bits(source_number[i], shift_distance)
    end for
    return source_number
  end if
  source_number = and_bits(source_number, #FFFFFFFF)
  if shift_distance = 0 then return source_number end if

  if shift_distance < 0 then
    source_number *= power(2, -shift_distance)
  else
    lSigned = 0
    -- Check for the sign bit so we don't propagate it.
    if and_bits(source_number, #80000000) then
      lSigned = 1
      source_number = and_bits(source_number, #7FFFFFFF)
    end if
    source_number /= power(2, shift_distance)
    if lSigned and shift_distance < 32 then
      -- Put back the sign bit now shifted
      source_number = or_bits(source_number, power(2, 31-shift_distance))
    end if
  end if

  return and_bits(source_number, #FFFFFFFF)
end function

-------------------------------------------------------------------------------

global function rotate_bits(object source_number, integer shift_distance)
--<function>
--<name>rotate_bits</name>
--<digest>Rotates the bits in the input value by the specified distance.</digest>
--<desc>
--</desc>
--<param>
--<type>object</type>
--<name>source_number</name>
--<desc></desc>
--</param>
--<param>
--<type>integer</type>
--<name>shift_distance</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- printf(1, "%08x\n", rotate_bits(#ABCD,  4))  --> D0000ABC
-- printf(1, "%08x\n", rotate_bits(#ABCD, -4))  --> 000ABCD0
--</example>
--<see_also>shift_bits()</see_also>
--</function>
  atom lTemp
  atom lSave
  integer lRest

  if sequence(source_number) then
    for i = 1 to length(source_number) do
      source_number[i] = rotate_bits(source_number[i], shift_distance)
    end for
    return source_number
  end if

  source_number = and_bits(source_number, #FFFFFFFF)
  if shift_distance = 0 then
    return source_number
  end if

  if shift_distance < 0 then
    lSave = not_bits(power(2, 32 + shift_distance) - 1)
    lRest = 32 + shift_distance
  else
    lSave = power(2, shift_distance) - 1
    lRest = shift_distance - 32
  end if

  lTemp = shift_bits(and_bits(source_number, lSave), lRest)
  source_number = shift_bits(source_number, shift_distance)
  return or_bits(source_number, lTemp)
end function


