-- OpenEuphoria std/mathcons.e and std/math.e
-- without following functions:
-- * larger_of  (see max_index in _search_.e)
-- * smaller_of (see min_index in _search_.e)
-- * ensure_in_range (see is_in_range)
-- * ensure_in_list (see find or find_in_array in _search_.e)
-- * binary functions (see _machine_.e)
-- * Eu 3.11 built-in functions

include _common_.e
include _types_.e

-------------------------------------------------------------------------------
-- Constants
-------------------------------------------------------------------------------

global constant
  PI        = 3.14159265358979323846,
--<constant>
--<name>PI</name>
--<value>3.14159265358979323846</value>
--<desc>ratio of a circle's circumference to it's diameter.</desc>
--</constant>
  QUARTPI   = 0.78539816339744830962,
--<constant>
--<name>QUARTPI</name>
--<value>0.78539816339744830962</value>
--<desc>Quarter of PI</desc>
--</constant>
  HALFPI    = 1.57079632679489661923,
--<constant>
--<name>HALFPI</name>
--<value>1.57079632679489661923</value>
--<desc>Half of PI</desc>
--</constant>
  TWOPI     = 6.28318530717958647692,
--<constant>
--<name>TWOPI</name>
--<value>6.28318530717958647692</value>
--<desc>Two times PI</desc>
--</constant>
  PISQR     = 9.86960440108935861883,
--<constant>
--<name>PISQR</name>
--<value>9.86960440108935861883</value>
--<desc>PI ^ 2</desc>
--</constant>
  INVSQ2PI  = 0.39894228040143367794,
--<constant>
--<name>INVSQ2PI</name>
--<value>0.39894228040143367794</value>
--<desc>1 / (sqrt(2PI))</desc>
--</constant>
  PHI       = 1.61803398874989484820,
--<constant>
--<name>PHI</name>
--<value>1.61803398874989484820</value>
--<desc>Golden Ratio = (1 + sqrt(5)) / 2</desc>
--</constant>
  E            = 2.71828182845904523536,
--<constant>
--<name>E</name>
--<value>2.71828182845904523536</value>
--<desc>Euler (e), base of the natural logarithm.</desc>
--</constant>
  LN2       = 0.69314718055994530941,
--<constant>
--<name>LN2</name>
--<value>0.69314718055994530941</value>
--<desc>ln(2) :: 2 = power(E, LN2)</desc>
--</constant>
  INVLN2    = 1.44269504088896340736,
--<constant>
--<name>INVLN2</name>
--<value>1.44269504088896340736</value>
--<desc>1 / (ln(2))</desc>
--</constant>
  LN10      = 2.30258509299404568401,
--<constant>
--<name>LN10</name>
--<value>2.30258509299404568401</value>
--<desc>ln(10) :: 10 = power(E, LN10)</desc>
--</constant>
  INVLN10   = 0.43429448190325182765,
--<constant>
--<name>INVLN10</name>
--<value>0.43429448190325182765</value>
--<desc>1 / ln(10)</desc>
--</constant>
  SQRT2     = 1.41421356237309504880,
--<constant>
--<name>SQRT2</name>
--<value>1.41421356237309504880</value>
--<desc>sqrt(2)</desc>
--</constant>
  HALFSQRT2 = 0.70710678118654752440,
--<constant>
--<name>HALFSQRT2</name>
--<value>0.70710678118654752440</value>
--<desc>sqrt(2)/ 2</desc>
--</constant>
  SQRT3     = 1.73205080756887729353,
--<constant>
--<name>SQRT3</name>
--<value>1.73205080756887729353</value>
--<desc>Square root of 3</desc>
--</constant>
  DEGREES_TO_RADIANS  = 0.01745329251994329576,
--<constant>
--<name>DEGREESTORADIANS</name>
--<value>0.01745329251994329576</value>
--<desc>Conversion factor: Degrees to Radians = PI / 180</desc>
--</constant>
  RADIANS_TO_DEGREES   = 57.29577951308232090712,
--<constant>
--<name>RADIANSTODEGREES</name>
--<value>57.29577951308232090712</value>
--<desc>Conversion factor: Radians to Degrees = 180 / PI</desc>
--</constant>
  EULERGAMMA  = 0.57721566490153286061,
--<constant>
--<name>EULERGAMMA</name>
--<value>0.57721566490153286061</value>
--<desc>Gamma (Euler Gamma)</desc>
--</constant>
  SQRTE        = 1.64872127070012814684
--<constant>
--<name>SQRTE</name>
--<value>1.64872127070012814684</value>
--<desc>sqrt(e)</desc>
--</constant>

global constant
  PINF     = 1E308 * 1000,
--<constant>
--<name>PINF</name>
--<value>1E308 * 1000</value>
--<desc>Positive Infinity</desc>
--</constant>
  MINF     = - PINF,
--<constant>
--<name>MINF</name>
--<value>- PINF</value>
--<desc>Negative Infinity</desc>
--</constant>
  SQRT5 = 2.23606797749978969641
--<constant>
--<name>SQRT5</name>
--<value>2.23606797749978969641</value>
--<desc>sqrt(5)</desc>
--</constant>

type trig_range(object x)
--  values passed to arccos and arcsin must be [-1,+1]
  if atom(x) then
    return x >= -1 and x <= 1
  else
    for i = 1 to length(x) do
      if not trig_range(x[i]) then return 0 end if
    end for
    return 1
  end if
end type

-------------------------------------------------------------------------------

global function abs(object a)
--<function>
--<name>abs</name>
--<digest>returns absolute value of a</digest>
--<desc>may be applied to an atom or to all elements of a sequence</desc>
--<param>
--<type>object</type>
--<name>a</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- x = abs({10.5, -12, 3})
-- x is {10.5, 12, 3}
-- i = abs(-4)
-- i is 4
--</example>
--<see_also>sign()</see_also>
--</function>
  object t

  if atom(a) then
    if a >= 0 then
      return a
    else
      return - a
    end if
  end if
  for i = 1 to length(a) do
    t = a[i]
    if atom(t) then
      if t < 0 then a[i] = - t end if
    else
      a[i] = abs(t)
    end if
  end for
  return a
end function

-------------------------------------------------------------------------------

global function sign(object a)
--<function>
--<name>sign</name>
--<digest>returns sign of a</digest>
--<desc>may be applied to an atom or to all elements of a sequence.</desc>
--<param>
--<type>object</type>
--<name>a</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- i = sign(5)
-- i is 1
-- i = sign(0)
-- i is 0
-- i = sign(-2)
-- i is -1
--</example>
--<see_also>abs(), compare()</see_also>
--</function>
  return (a > 0) - (a < 0)
end function

-------------------------------------------------------------------------------

global function max(object a)
--<function>
--<name>max</name>
--<digest>returns maximal value of a</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>object</type>
--<name>a</name>
--<desc></desc>
--</param>
--<return>
-- object
--</return>
--<example>
-- a = max({10,15.4,3})
-- a is 15.4
--</example>
--<see_also>min()</see_also>
--</function>
  atom b, c

  if atom(a) then return a end if
  b = MINF
  for i = 1 to length(a) do
    c = max(a[i])
    if c > b then b = c end if
  end for
  return b
end function

-------------------------------------------------------------------------------

global function min(object a)
--<function>
--<name>min</name>
--<digest>returns minimal value of a</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>object</type>
--<name>a</name>
--<desc></desc>
--</param>
--<return>
-- object
--</return>
--<example>
-- a = min({10,15.4,3})
-- a is 3
--</example>
--<see_also>max()</see_also>
--</function>
  atom b, c

  if atom(a) then return a end if
  b = PINF
  for i = 1 to length(a) do
    c = min(a[i])
    if c < b then b = c end if
  end for
  return b
end function

-------------------------------------------------------------------------------

global function is_in_range(atom item, atom from, atom upto)
--<function>
--<name>is_in_range</name>
--<digest>checks if item fills in a range</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>item</name>
--<desc>item to check</desc>
--</param>
--<param>
--<type>atom</type>
--<name>from</name>
--<desc>minimal value</desc>
--</param>
--<param>
--<type>atom</type>
--<name>upto</name>
--<desc>maximal value</desc>
--</param>
--<return>
-- 1 if within range, 0 if not
--</return>
--<example>
-- i = is_in_range(3.14, 3, 4)
-- i is 1
-- i = is_in_range(3.14, 2, 3)
-- i is 0
--</example>
--<see_also></see_also>
--</function>
  if compare(item, from) < 0 then return 0 end if
  if compare(item, upto) > 0 then return 0 end if
  return 1
end function

-------------------------------------------------------------------------------

global function mod(object x, object y)
--<function>
--<name>mod</name>
--<digest>returns remainder of a division using floored division</digest>
--<desc>
-- when the operands' signs are different rounds dividend/divisor away from zero
-- whereas remainder() rounds towards zero.
--</desc>
--<param>
--<type>integer</type>
--<name>x</name>
--<desc>dividend</desc>
--</param>
--<param>
--<type>integer</type>
--<name>y</name>
--<desc>divisor</desc>
--</param>
--<return>
-- object: remainder of the division
--</return>
--<example>
-- a = mod(9, 4)
-- a is 1
-- s = mod({81, -3.5, -9, 5.5}, {8, -1.7, 2, -4})
-- s is {1,-0.1,1,-2.5}
-- s = mod({17, 12, 34}, 16)
-- s is {1, 12, 2}
-- s = mod(16, {2, 3, 5})
-- s is {0, 1, 1}
--</example>
--<see_also>
--</see_also>
--</function>
  if equal(sign(x), sign(y)) then
    return remainder(x,y)
  end if
  return x - y * floor(x / y)
end function

-------------------------------------------------------------------------------

global function trunc(object x)
--<function>
--<name>trunc</name>
--<digest>returns integer part of an object</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>object</type>
--<name>x</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- a = trunc(9.4)
-- a is 9
-- s = trunc({81, -3.5, -9.999, 5.5})
-- s is {81,-3, -9, 5}
--</example>
--<see_also>frac()</see_also>
--</function>
  return sign(x) * floor(abs(x))
end function

-------------------------------------------------------------------------------

global function frac(object x)
--<function>
--<name>frac</name>
--<digest>returns fractional part of an object</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>object</type>
--<name>x</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- a = frac(9.4)
-- a is 0.4
-- s = frac({81, -3.5, -9.999, 5.5})
-- s is {0, -0.5, -0.999, 0.5}
--</example>
--<see_also>trunc()</see_also>
--</function>
  object temp

  temp = abs(x)
  return sign(x) * (temp - floor(temp))
end function

-------------------------------------------------------------------------------

global function ceil(object a)
--<function>
--<name>ceil</name>
--<digest></digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>object</type>
--<name>a</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- i = ceil(15.2)
-- i is 16
-- s = ceil({3.14, 1.72})
-- s is {4, 2}
--</example>
--<see_also>floor()</see_also>
--</function>
  return -floor(-a)
end function

-------------------------------------------------------------------------------

global function intdiv(object a, object b)
--<function>
--<name>intdiv</name>
--<digest>returns result of an integer division</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>integer</type>
--<name>a</name>
--<desc>dividend</desc>
--</param>
--<param>
--<type>integer</type>
--<name>b</name>
--<desc>divisor</desc>
--</param>
--<return>
--</return>
--<example>
-- a = intdiv(9, 4)
-- a is 3
-- s = intdiv({81, -3.5, -9, 5.5}, {8, -1.7, 2, -4})
-- s is {11,-3,-5,2}
-- s = intdiv({17, 12, 34}, 16)
-- s is {2,1,3}
-- s = intdiv(16, {2, 3, 5})
-- s is {8,6,4}
--</example>
--<see_also>remainder(), mod()</see_also>
--</function>
  return sign(a)*ceil(abs(a)/abs(b))
end function

-------------------------------------------------------------------------------

global function round(object a, sequence optional)
--<function>
--<name>round</name>
--<digest>rounds an object to a specified precision</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>object</type>
--<name>a</name>
--<desc></desc>
--</param>
--<param>
--<type>sequence</type>
--<name>optional</name>
-- optional "precision" indicates number of digits to round behind dot. defaults to 0
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- round(5.2, {})
-- 5
-- round({4.12, 4.67, -5.8, -5.21}, {{"precision",10}})
-- {4.1, 4.7, -5.8, -5.2}
-- round(12.2512, {{"precision",100}})
-- 12.25
--</example>
--<see_also>
--</see_also>
--</function>
  integer len
  sequence s
  object t, u, precision

  precision = abs(get_option("precision", optional, 1))
  if atom(a) then
    if atom(precision) then
      return floor(0.5 + (a * precision )) / precision
    end if
    len = length(precision)
    s = repeat(0, len)
    for i = 1 to len do
      t = precision[i]
      if atom (t) then
        s[i] = floor( 0.5 + (a * t)) / t
      else
        s[i] = round(a, t)
      end if
    end for
    return s
  elsif atom(precision) then
    len = length(a)
    s = repeat(0, len)
    for i = 1 to len do
      t = a[i]
      if atom(t) then
        s[i] = floor(0.5 + (t * precision)) / precision
      else
        s[i] = round(t, precision)
      end if
    end for
    return s
  end if
  len = length(a)
  if len != length(precision) then
    error_message("The lengths of the two supplied sequences do not match.", 1)
  end if
  s = repeat(0, len)
  for i = 1 to len do
    t = precision[i]
    if atom(t) then
      u = a[i]
      if atom(u) then
        s[i] = floor(0.5 + (u * t)) / t
      else
        s[i] = round(u, t)
      end if
    else
      s[i] = round(a[i], t)
    end if
  end for
  return s
end function

-------------------------------------------------------------------------------

global function rad2deg (object x)
--<function>
--<name>rad2deg</name>
--<digest>Convert an angle measured in radians to an angle measured in degrees</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>object</type>
--<name>x</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- x = rad2deg(3.385938749)
-- x is 194
--</example>
--<see_also>
--</see_also>
--</function>
  return x * RADIANS_TO_DEGREES
end function

-------------------------------------------------------------------------------

global function deg2rad (object x)
--<function>
--<name>deg2rad</name>
--<digest>Convert an angle measured in degrees to an angle measured in radians</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>object</type>
--<name>x</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- x = deg2rad(194)
-- x is 3.385938749
--</example>
--<see_also>
--</see_also>
--</function>
  return x * DEGREES_TO_RADIANS
end function

-------------------------------------------------------------------------------

global function log10(object x1)
--<function>
--<name>log10</name>
--<digest>base 10 logarithm of a number.</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>object</type>
--<name>x1</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- a = log10(12)
-- a is 1.079181246 (use pretty_print instead of printf to get more digits)
-- s = log10({12, 15})
-- s is {1.079181246, 1.176091259}
--</example>
--<see_also>
--</see_also>
--</function>
  return log(x1) * INVLN10
end function

-------------------------------------------------------------------------------

global function exp(object x)
--<function>
--<name>exp</name>
--<digest>Computes some power of E.</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>atom</type>
--<name>x</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- x = exp(5.4)
-- x is 221.4064162
-- s =exp({1, 5.4}), {0})
-- s is {2.718281828, 221.4064162}
--</example>
--<see_also>
--</see_also>
--</function>
  return power(E, x)
end function

-------------------------------------------------------------------------------

global function fib(integer i)
--<function>
--<name>fib</name>
--<digest>Computes the Fibonacci Number of the given value</digest>
--<desc>
-- fib(0) = 0
-- fib(1) = 1
-- fib(n+2) = fib(n+1) + fib(n) if n > 1
--</desc>
--<param>
--<type>integer</type>
--<name>i</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- i = fib(6)
-- i is 8
--</example>
--<see_also>
--</see_also>
--</function>
  return floor((power( PHI, i) / SQRT5) + 0.5)
end function

-------------------------------------------------------------------------------

global function cosh(object a)
--<function>
--<name>cosh</name>
--<digest>hyperbolic cosine of an object.</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>object</type>
--<name>a</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>? cosh(LN2) -- prints out 1.25--</example>
--<see_also>sinh(), arccosh()</see_also>
--</function>
  return (exp(a)+exp(-a))/2
end function

-------------------------------------------------------------------------------

global function sinh(object a)
--<function>
--<name>sinh</name>
--<digest>hyperbolic sine of an object.</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>object</type>
--<name>a</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>? sinh(LN2) -- prints out 0.75</example>
--<see_also>cosh(), arcsinh()</see_also>
--</function>
  return (exp(a)-exp(-a))/2
end function

-------------------------------------------------------------------------------

global function tanh(object a)
--<function>
--<name>tanh</name>
--<digest>hyperbolic tangent of an object.</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>object</type>
--<name>a</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>? tanh(LN2) -- prints out 0.6</example>
--<see_also>cosh(), sinh(), arctanh()</see_also>
--</function>
  return sinh(a)/cosh(a)
end function

-------------------------------------------------------------------------------

global function arcsinh(object a)
--<function>
--<name>arcsinh</name>
--<digest>reverse hyperbolic sine of an object.</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>object</type>
--<name>a</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>? arcsinh(1) -- prints out 0.881373587</example>
--<see_also>
--</see_also> arccosh(), sinh()</function>
  return log(a+sqrt(1+a*a))
end function

-------------------------------------------------------------------------------

type not_below_1(object x)
  if atom(x) then
    return x>=1.0
  end if
  for i=1 to length(x) do
    if not not_below_1(x[i]) then
      return 0
    end if
  end for
  return 1
end type

-------------------------------------------------------------------------------

global function arccosh(not_below_1 a)
--<function>
--<name>arccosh</name>
--<digest>reverse hyperbolic cosine of an object.</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>not_below_1</type>
--<name>a</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>? arccosh(1) -- prints out 0</example>
--<see_also>arcsinh(), cosh()</see_also>
--</function>
  return log(a+sqrt(a*a-1))
end function

-------------------------------------------------------------------------------

type abs_below_1(object x)
  if atom(x) then
    return x>-1.0 and x<1.0
  end if
  for i=1 to length(x) do
    if not abs_below_1(x[i]) then
      return 0
    end if
  end for
  return 1
end type

-------------------------------------------------------------------------------

global function arctanh(abs_below_1 a)
--<function>
--<name>arctanh</name>
--<digest>reverse hyperbolic tangent of an object.</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>abs_below_1</type>
--<name>a</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>? arctanh(1/2) -- prints out 0.5493061443</example>
--<see_also>arcsinh(), cosh()</see_also>
--</function>
  return log((1+a)/(1-a))/2
end function

-------------------------------------------------------------------------------

global function sum(object a)
--<function>
--<name>sum</name>
--<digest>sum of all atoms in the argument.</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>object</type>
--<name>a</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- a = sum({10, 20, 30})
-- a is 60
-- a = sum({10.5, {11.2} , 8.1})
-- a is 29.8
--</example>
--<see_also>product()</see_also>
--</function>
  atom b

  if atom(a) then return a end if
  b = 0
  for i = 1 to length(a) do
    if atom(a[i]) then
      b += a[i]
    else
      b += sum(a[i])
    end if
  end for
  return b
end function

-------------------------------------------------------------------------------

global function product(object a)
--<function>
--<name>product</name>
--<digest>product of all the atom in the argument.</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>object</type>
--<name>a</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- a = product({10, 20, 30})
-- a is 6000
-- a = product({10.5, {11.2} , 8.1})
-- a is 952.56
--</example>
--<see_also>sum()</see_also>
--</function>
  atom b

  if atom(a) then return a end if
  b = 1
  for i = 1 to length(a) do
    if atom(a[i]) then
      b *= a[i]
    else
      b *= product(a[i])
    end if
  end for
  return b
end function

-------------------------------------------------------------------------------

global function gcd(atom p, atom q)
--<function>
--<name>gcd</name>
--<digest>greater common divisor of to atoms</digest>
--<desc>
-- largest value that evenly divides into both parameters
-- zero is ignored
-- </desc>
--<param>
--<type>atom</type>
--<name>p</name>
--<desc></desc>
--</param>
--<param>
--<type>atom</type>
--<name>q</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- ? gcd(76.3, -114) --> 38
-- ? gcd(0, -114) --> 114
-- ? gcd(0, 0) --> 0 (This is often regarded as an error condition)
--</example>
--<see_also>
--</see_also>
--</function>
  atom r

  -- Both arguments must be positive.
  if p < 0 then p = -p end if
  if q < 0 then q = -q end if

  -- Strip off any fractional part.
  p = floor(p)
  q = floor(q)

  -- Ensure that 'p' is not smaller than 'q'
  if p < q then
    r = p
    p = q
    q = r
  end if

  -- Special case.
  if q = 0 then return p end if

  -- repeat until I get a remainder less than 2.
  r = remainder(p, q)
  while r > 1 do
    -- set up next cycle using denominator and remainder from previous cycle.
    p = q
    q = r
    -- get remainder after dividing p by q
    r = remainder(p, q)
  end while

  if r = 1 then
    return 1
  else
    return q
  end if
end function

-------------------------------------------------------------------------------

global function approx(object p, object q, sequence optional)
--<function>
--<name>approx</name>
--<digest>Compares two (sets of) numbers based on approximate equality.</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>object</type>
--<name>p</name>
--<desc></desc>
--</param>
--<param>
--<type>object</type>
--<name>q</name>
--<desc></desc>
--</param>
--<param>
--<type>sequence</type>
--<name>optional</name>
--<desc>
-- optional "epsilon" defines the amount of inequality allowed.
-- This must be a positive value. Default is 0.005
--</desc>
--</param>
--<return>
--</return>
--<example>
-- ? approx(10, 33.33 * 30.01 / 100, {})
-- prints 0 because 10 and 10.002333 are within 0.005 of each other
-- ? approx(10, 10.001, {})
-- prints 0 because 10 and 10.001 are within 0.005 of each other
-- ? approx(10, {10.001,9.999, 9.98, 10.04}, {})
-- prints {0,0,1,-1}
-- ? approx({10.001,9.999, 9.98, 10.04}, 10, {})
-- prints {0,0,-1,1}
-- ? approx({10.001,{9.999, 10.01}, 9.98, 10.04}, {10.01,9.99, 9.8, 10.4}, {})
-- prints {-1,{1,1},1,-1}
-- ? approx(23, 32, {{"epsilon",10}})
-- prints 0 because 23 and 32 are within 10 of each other
--</example>
--<see_also>
--</see_also>
--</function>
  atom epsilon
  
  epsilon = get_option("epsilon", optional, 0.005)
  if sequence(p) then
    if sequence(q) then
      if length(p) != length(q) then
        error_message("approx(): Sequence arguments must be the same length", 1)
      end if
      for i = 1 to length(p) do
        p[i] = approx(p[i], q[i], optional)
      end for
      return p
    else
      for i = 1 to length(p) do
        p[i] = approx(p[i], q, optional)
      end for
      return p
    end if
  elsif sequence(q) then
    for i = 1 to length(q) do
      q[i] = approx(p, q[i], optional)
    end for
    return q
  else
printf(1, "p = %d, q = %d, epsilon = %g\n", {p, q, epsilon})
printf(1, "p = %d, q + epsilon = %g, q - epsilon = %g\n", {p, q + epsilon, q - epsilon})
    if p > (q + epsilon) then return 1 end if
    if p < (q - epsilon) then return -1 end if
    return 0
  end if
end function

-------------------------------------------------------------------------------

global function is_powerof2(object p)
--<function>
--<name>is_powerof2</name>
--<digest>checks if object is power of 2</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>object</type>
--<name>p</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
--for i = 1 to 5 do
--  ? {i, powof2(i)}
--end for
-- output ...
-- {1,1}
-- {2,1}
-- {3,0}
-- {4,1}
-- {5,0}
--</example>
--<see_also>
--</see_also>
--</function>
  return not (and_bits(p, p-1))
end function

-------------------------------------------------------------------------------

global function is_even(object test_object)
--<function>
--<name>is_even</name>
--<digest>Test if the supplied Euphoria object is even.</digest>
--<desc>may be applied to an atom or to a sequence</desc>
--<param>
--<type>object</type>
--<name>test_object</name>
--<desc></desc>
--</param>
--<return>
--</return>
--<example>
-- ? is_even(2) --> 1
-- ? is_even(1) --> 0
-- ? is_even(-2) --> 1
-- ? is_even(-1) --> 0
-- ? is_even({-1,0,2})  --> {0, 1, 1}
-- ? is_even(3.4) --> 0
-- ? is_even({{1,2,3}, {{4,5},6,{7,8}},9}) --> {{0,1,0},{{1,0},1,{0,1}},0}
--</example>
--<see_also>
--</see_also>
--</function>
  if atom(test_object) then
    if integer(test_object) then
      return (and_bits(test_object, 1) = 0)
    end if
    return 0
  end if
  for i = 1 to length(test_object) do
    test_object[i] = is_even(test_object[i])
  end for

  return test_object
end function

