--****
-- == Pipe Input/Output
--
-- <<LEVELTOC level=2 depth=4>>
--
-- === Notes
-- Due to a bug, Euphoria does not handle STDERR properly
-- STDERR cannot captured for Euphoria programs (other programs will work fully)
-- The IO functions currently work with file handles, a future version might wrap them in streams
-- so that they can be used directly alongside other file/socket/other-streams with a
-- stream_select() function.
--

include _dll_.e
include _machine_.e
include _w32constants_.e
include _w32routines_.e
include _w32errors_.e as err

constant
  PIPE_WRITE_HANDLE = 1, PIPE_READ_HANDLE=2,
  HANDLE_FLAG_INHERIT=1,
  STARTF_USESHOWWINDOW = 1,
  STARTF_USESTDHANDLES = 256,
  FAIL = 0

constant
  SA_SIZE = 12,
  SUIdwFlags = 44, 
  SUIhStdInput = 56, 
  STARTUPINFO_SIZE = 68,
  PROCESS_INFORMATION_SIZE = 16

--****
-- === Accessor Constants

global constant
  --** Child processes standard input
  STDIN=1,
  --** Child processes standard output
  STDOUT=2,
  --** Child processes standard error
  STDERR=3,
  --** Process ID
  PID=4
--<constant>
--<name>STDIN</name>
--<value>1</value>
--<desc></desc>
--</constant>
--<constant>
--<name>STDOUT</name>
--<value>2</value>
--<desc></desc>
--</constant>
--<constant>
--<name>STDERR</name>
--<value>3</value>
--<desc></desc>
--</constant>
--<constant>
--<name>PID</name>
--<value>4</value>
--<desc></desc>
--</constant>

--
global constant
  --** Set of pipes that are for the use of the parent
  PARENT=1,
  --** Set of pipes that are given to the child - should not be used by the parent
  CHILD=2
--<constant>
--<name>PARENT</name>
--<value>1</value>
--<desc></desc>
--</constant>
--<constant>
--<name>CHILD</name>
--<value>2</value>
--<desc></desc>
--</constant>

error_code = 0

--------------------------------------------------------------------------------

global type process(object o)
--<eu_type>
--<name>process</name>
--<desc></desc>
--<param>
--<type>object</type>
--<name>o</name>
--<desc></desc>
--</param>
--</eu_type>
  if atom(o) then
    return 0
  end if

  if length(o) != 4 then
    return 0
  end if

  return 1
end type

--------------------------------------------------------------------------------

integer xCloseHandle
xCloseHandle = 0

global function close_pipe(atom fd)
--<function>
--<name>close_pipe</name>
--<digest>Close handle</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>fd</name>
--<desc></desc>
--</param>
--<return>
-- integer: 0 on success, -1 on failure
--</return>
--<example>
-- integer status
-- status = pipeio:close(p[STDIN])
--</example>
--<see_also>
--</see_also>
--</function>
  atom ret

  ret = exec_w32func(xCloseHandle, {fd}, "CloseHandle")
  if ret = FAIL then
    error_code = err:GetLastError()
    error_msg  = err:FormatSystemError(error_code)
    log_printf("Error %d: %s\n", {error_code, error_msg})
    return -1
  end if
  
  return 0
end function

--------------------------------------------------------------------------------

integer xTerminateProcess
xTerminateProcess = 0

global procedure kill_pipe(process p, atom signal)
--<procedure>
--<name>kill_pipe</name>
--<digest>Close pipes and kill process p with signal</digest>
--<desc>
-- Comments:
--   Signal is ignored on Windows.
--</desc>
--<param>
--<type>process</type>
--<name>p</name>
--<desc></desc>
--</param>
--<param>
--<type>atom</type>
--<name>signal</name>
--<desc></desc>
--</param>
--<example>
-- kill_pipe(p)
--</example>
--<see_also>
--</see_also>
--</procedure>
  void = close_pipe(p[STDIN])
  void = close_pipe(p[STDOUT])
  void = close_pipe(p[STDERR])
  
  void = exec_w32func(xTerminateProcess, {p[PID],signal and 0},
                   "TerminateProcess")
end procedure

--------------------------------------------------------------------------------

integer xCreatePipe
xCreatePipe = 0

function os_pipe()
  sequence handles
  atom ret, psaAttrib, phWriteToPipe, phReadFromPipe
  
  psaAttrib = allocate(SA_SIZE + 2 * size_of( C_POINTER ) )
  poke4( psaAttrib, SA_SIZE )
  poke_pointer(psaAttrib + size_of( C_POINTER),{0,1})
  
  phWriteToPipe = psaAttrib + SA_SIZE
  phReadFromPipe = psaAttrib + SA_SIZE + size_of( C_POINTER )
  
  ret = exec_w32func(xCreatePipe, {phReadFromPipe, phWriteToPipe, psaAttrib, 0},
                  "CreatePipe")
  handles = peek_pointer({phWriteToPipe,2})
  
  free(psaAttrib)
  
  if ret = FAIL then
    error_code = err:GetLastError()
    error_msg  = err:FormatSystemError(error_code)
    log_printf("Error %d: %s\n", {error_code, error_msg})
    return -1
  end if
  
  return handles
end function

--------------------------------------------------------------------------------

integer xReadFile
xReadFile = 0

global function read_pipe(atom fd, integer bytes)
--<function>
--<name>read_pipe</name>
--<digest>Reads a sequence from a named pipe</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>fd</name>
--<desc></desc>
--</param>
--<param>
--<type>integer</type>
--<name>bytes</name>
--<desc></desc>
--</param>
--<return>
-- sequence: either data, an empty sequence on EOF or an error code.
--</return>
--<example>
-- sequence data
-- data = read_pipe(p[STDOUT], 256)
--</example>
--<see_also>get_bytes()</see_also>
--</function>
  sequence data
  atom ret, ReadCount, buf, pReadCount

  if bytes=0 then return "" end if

  buf = allocate(bytes)
  
  pReadCount=allocate(4)
  ret = exec_w32func(xReadFile, {fd, buf, bytes, pReadCount, 0}, "ReadFile")
  ReadCount=peek4u(pReadCount)
  free(pReadCount)
  
  if ret = FAIL then
    error_code = err:GetLastError()
    error_msg  = err:FormatSystemError(error_code)
    log_printf("Error %d: %s\n", {error_code, error_msg})
    free(buf)
    return ""
  end if
  
  data = peek({buf, ReadCount})
  free(buf)
  
  return data
end function

--------------------------------------------------------------------------------

integer xWriteFile
xWriteFile = 0

global function write_pipe(atom fd, sequence str)
--<function>
--<name>write_pipe</name>
--<digest>Writes a string to a named pipe</digest>
--<desc>
--</desc>
--<param>
--<type>atom</type>
--<name>fd</name>
--<desc></desc>
--</param>
--<param>
--<type>sequence</type>
--<name>str</name>
--<desc></desc>
--</param>
--<return>
-- integer: number of bytes written, or -1 on error
--</return>
--<example>
-- integer bytes_written
-- bytes_written = write_pipe(p[STDIN],"Hello World!")
--</example>
--<see_also>
--</see_also>
--</function>
  atom buf, ret, WrittenCount, pWrittenCount
  
  buf = allocate_string(str)
  pWrittenCount=allocate(4)
  ret = exec_w32func(xWriteFile, {fd, buf, length(str), pWrittenCount, 0},
                  "WriteFile")
  WrittenCount = peek4u(pWrittenCount)
  free(pWrittenCount)
  
  free(buf)
  
  if ret = FAIL then
    error_code = err:GetLastError()
    error_msg  = err:FormatSystemError(error_code)
    log_printf("Error %d: %s\n", {error_code, error_msg})
    return -1
  end if

  return WrittenCount
end function

--------------------------------------------------------------------------------

global function error_no()
--<function>
--<name>error_no</name>
--<digest>Get error no from last call to a pipe function</digest>
--<desc>
-- Comments:
-- Value returned will be OS-specific, and is not always set on Windows at least
--</desc>
--<return>
--</return>
--<example>
-- integer error
-- error = error_no()
--</example>
--<see_also>
--</see_also>
--</function>
  return error_code
end function

--------------------------------------------------------------------------------

integer xGetStdHandle
xGetStdHandle = 0

function GetStdHandle(atom device)
  return exec_w32func(xGetStdHandle, {device}, "GetStdHandle")
end function

--------------------------------------------------------------------------------

integer xSetHandleInformation
xSetHandleInformation = 0

function SetHandleInformation(atom hObject, atom dwMask, atom dwFlags)
  return exec_w32func(xSetHandleInformation, {hObject, dwMask, dwFlags},
                    "SetHandleInformation")
end function

--------------------------------------------------------------------------------

procedure CloseAllHandles(sequence handles)
  for i = 1 to length(handles) do
    void = close_pipe(handles[i])
  end for
end procedure

--------------------------------------------------------------------------------

integer xCreateProcess
xCreateProcess = 0

function CreateProcess(sequence CommandLine, sequence StdHandles)
  object fnVal
  atom pPI, pSUI, pCmdLine
  sequence ProcInfo

  pCmdLine = allocate_string(CommandLine)
  
  pPI = allocate(PROCESS_INFORMATION_SIZE)
  mem_set(pPI,0,PROCESS_INFORMATION_SIZE)
  
  pSUI = allocate(STARTUPINFO_SIZE)
  mem_set( pSUI, 0, STARTUPINFO_SIZE)
  poke4( pSUI, STARTUPINFO_SIZE)
  poke4( pSUI + SUIdwFlags, or_bits( STARTF_USESTDHANDLES,
                                      STARTF_USESHOWWINDOW ) )
  poke_pointer( pSUI + SUIhStdInput, StdHandles)
  
  fnVal = exec_w32func(xCreateProcess, {0, pCmdLine, 0, 0, 1, 0, 0, 0, pSUI, pPI},
                    "CreateProcessA")
  free(pCmdLine)
  free(pSUI)
  ProcInfo = peek4u({pPI,4})
  free(pPI)
  if not fnVal then
    return 0
  end if
  return ProcInfo
end function

-- Function not implemented (not needed)
-- iGetExitCodeProcess=define_c_func(kernel32,"GetExitCodeProcess",
--                                   {C_UINT,C_POINTER},C_INT),

--------------------------------------------------------------------------------

global function create_pipe()
--<function>
--<name>create_pipe</name>
--<digest>Create pipes for inter-process communication</digest>
--<desc>
--</desc>
--<return>
-- sequence: process handles { {parent side pipes},{child side pipes} }
--</return>
--<example>
-- object p
-- p = exec_pipe("dir", create_pipe())
--</example>
--<see_also>
--</see_also>
--</function>
  atom hChildStdInRd,hChildStdOutWr, hChildStdErrWr,
       hChildStdInWr, hChildStdOutRd,hChildStdErrRd

  object StdInPipe, StdOutPipe, StdErrPipe

  StdInPipe = {}
  StdOutPipe = {}
  StdErrPipe = {}
  
  StdInPipe = os_pipe()
  if atom(StdInPipe) then return -1 end if
  hChildStdInRd = StdInPipe[PIPE_READ_HANDLE]
  hChildStdInWr = StdInPipe[PIPE_WRITE_HANDLE]
  
  StdOutPipe = os_pipe()
  if atom(StdOutPipe) then 
    CloseAllHandles(StdInPipe)
    return -1
  end if
  hChildStdOutWr = StdOutPipe[PIPE_WRITE_HANDLE]
  hChildStdOutRd = StdOutPipe[PIPE_READ_HANDLE]
  
  StdErrPipe = os_pipe()
  if atom(StdErrPipe) then
     CloseAllHandles(StdErrPipe & StdOutPipe)
     return -1
  end if
  hChildStdErrWr = StdErrPipe[PIPE_WRITE_HANDLE]
  hChildStdErrRd = StdErrPipe[PIPE_READ_HANDLE]

  void = SetHandleInformation(StdInPipe[PIPE_WRITE_HANDLE],HANDLE_FLAG_INHERIT,0)
  void = SetHandleInformation(StdOutPipe[PIPE_READ_HANDLE],HANDLE_FLAG_INHERIT,0)
  void = SetHandleInformation(StdErrPipe[PIPE_READ_HANDLE],HANDLE_FLAG_INHERIT,0)

  return {{hChildStdInWr, hChildStdOutRd, hChildStdErrRd},
           {hChildStdInRd, hChildStdOutWr, hChildStdErrWr}}
  
end function

--------------------------------------------------------------------------------

global function exec_pipe(sequence cmd, sequence pipe)
--<function>
--<name>exec_pipe</name>
--<digest>Open process with command line cmd</digest>
--<desc>
--</desc>
--<param>
--<type>sequence</type>
--<name>cmd</name>
--<desc></desc>
--</param>
--<param>
--<type>sequence</type>
--<name>pipe</name>
--<desc></desc>
--</param>
--<return>
-- sequence: process handles { [[:PID]], [[:STDIN]], [[:STDOUT]], [[:STDERR]] }
--</return>
--<example>
-- object p
-- p = exec_pipe("dir", create_pipe())
--</example>
--<see_also>
--</see_also>
--</function>
  object fnVal
  atom hChildStdInRd,hChildStdOutWr, hChildStdErrWr,
       hChildStdInWr, hChildStdOutRd,hChildStdErrRd
  atom hChildProcess

  hChildStdInWr = pipe[1][1]
  hChildStdOutRd = pipe[1][2]
  hChildStdErrRd = pipe[1][3]
  hChildStdInRd = pipe[2][1]
  hChildStdOutWr = pipe[2][2]
  hChildStdErrWr = pipe[2][3]

  fnVal = CreateProcess(cmd,{hChildStdInRd,hChildStdOutWr,hChildStdErrWr})
  if atom(fnVal) then return -1 end if
  hChildProcess = fnVal[1]
  void = close_pipe(fnVal[2])
  void = close_pipe(hChildStdInRd)
  void = close_pipe(hChildStdOutWr)
  void = close_pipe(hChildStdErrWr)
  return {hChildStdInWr, hChildStdOutRd, hChildStdErrRd, hChildProcess}

end function

------------------------------------------------------------------------------

global function exec_command(sequence cmd)
--<function>
--<name>exec_command</name>
--<digest>Executes a command in a named pipe</digest>
--<desc>
--</desc>
--<param>
--<type>sequence</type>
--<name>cmd</name>
--<desc>Command to execute</desc>
--</param>
--<return>
-- sequence: text output by the command
--</return>
--<example>
--</example>
--<see_also>
--</see_also>
--</function>
  sequence s
  object z, p, c

  s = ""
  z = create_pipe()
  p = exec_pipe(cmd, z)
  if atom(p) then
    kill_pipe(p, 15)
    return -1
  end if
  c = read_pipe(p[STDOUT], 256)
  while sequence(c) and length(c) do
    s &= c
    if atom(c) then
      kill_pipe(p, 15)
      return -1
    end if
    c = read_pipe(p[STDOUT], 256)
  end while
  --Close pipes and make sure process is terminated
  kill_pipe(p, 15)
  return s
end function
